@@ -572,8 +572,6 @@ GtkWidget *get_dial_box(session_t *session) {
   gtk_signal_connect(GTK_OBJECT(session->pick_up_button), "clicked",
 		     GTK_SIGNAL_FUNC(gtk_handle_pick_up_button),
 		     (gpointer) session);
-  gtk_widget_set_size_request(session->pick_up_button, 82, -1);
-  /* XXX: set automatically to maximum possibly needed size */
   gtk_widget_show(session->pick_up_button);
 
   /* activate dial button when pressing enter in entry widget */
@@ -588,8 +586,7 @@ GtkWidget *get_dial_box(session_t *session) {
 
   /* pick up button hbox */
   label_hbox = gtk_hbox_new(FALSE, 0);
-  gtk_container_add(GTK_CONTAINER(session->pick_up_button),
-		    label_hbox);
+  gtk_container_add(GTK_CONTAINER(session->pick_up_button), label_hbox);
   gtk_widget_show(label_hbox);
 
   /* pick up button symbol */
@@ -606,7 +603,7 @@ GtkWidget *get_dial_box(session_t *session) {
   /* pick up button label */
   session->pick_up_label = gtk_label_new(NULL);
   gtk_box_pack_start(GTK_BOX(label_hbox), session->pick_up_label,
-		     TRUE, FALSE, 3); /* expand but fill up outside label */
+		     TRUE, FALSE, 16); /* expand but fill up outside label */
   gtk_widget_show(session->pick_up_label);
 
    /* hang up button */
@@ -615,8 +612,6 @@ GtkWidget *get_dial_box(session_t *session) {
   gtk_signal_connect(GTK_OBJECT(session->hang_up_button), "clicked",
 		     GTK_SIGNAL_FUNC(gtk_handle_hang_up_button),
 		     (gpointer) session);
-  gtk_widget_set_size_request(session->hang_up_button, 82, -1);
-  /* XXX: set automatically to maximum possibly needed size */
   gtk_widget_show(session->hang_up_button);
 
   /* pick up button hbox */
@@ -641,7 +636,7 @@ GtkWidget *get_dial_box(session_t *session) {
   /* hang up button label */
   session->hang_up_label = gtk_label_new(NULL);
   gtk_box_pack_start(GTK_BOX(label_hbox), session->hang_up_label,
-		     TRUE, FALSE, 3); /* expand but fill up outside label */
+		     TRUE, FALSE, 16); /* expand but fill up outside label */
   gtk_widget_show(session->hang_up_label);
   
   return frame;
