@@ -32,12 +32,16 @@
 #ifdef HAVE_UNISTD_H
   #include <unistd.h>
 #endif
+#include <fcntl.h>
 #include <math.h>
 #include <time.h>
 
 /* GTK */
 #include <gtk/gtk.h>
 
+/* libsndfile */
+#include <sndfile.h>
+
 /* own header files */
 #include "globals.h"
 #include "session.h"
@@ -63,7 +67,8 @@ struct state_data_t state_data[STATE_NUMBER] = {
 {N_("RING"),          N_("Answer"), 1,N_("Reject"), 1},/* STATE_RINGING_QUIET */
 {N_("Dialing"),       N_("Pick up"),0,N_("Cancel"), 1},/* STATE_DIALING       */
 {N_("B-Channel open"),N_("Pick up"),0,N_("Hang up"),1},/* STATE_CONVERSATION  */
-{N_("Setup"),         N_("Pick up"),0,N_("Hang up"),0} /* STATE_SERVICE       */
+{N_("Setup"),         N_("Pick up"),0,N_("Hang up"),0},/* STATE_SERVICE       */
+{N_("Playback"),      N_("Pick up"),0,N_("Stop")   ,1} /* STATE_PLAYBACK      */
 };
 
 /*
@@ -487,21 +492,27 @@ void session_io_handlers_start(session_t *session) {
   if (!(session->option_release_devices &&
 	(session->state == STATE_READY ||
 	 session->state == STATE_RINGING_QUIET))) {
-    session->gdk_audio_input_tag = gdk_input_add(session->audio_fd_in,
-						 GDK_INPUT_READ,
-						 gdk_handle_audio_input,
-						 (gpointer) session);
+    session->gtk_audio_input_tag = gtk_input_add_full(session->audio_fd_in,
+						      GDK_INPUT_READ,
+						      gtk_handle_audio_input,
+						      NULL,
+						      (gpointer) session,
+						      NULL);
   }
-  session->gdk_isdn_input_tag = gdk_input_add(session->isdn_fd,
-					      GDK_INPUT_READ,
-					      gdk_handle_isdn_input,
-					      (gpointer) session);
+  session->gtk_isdn_input_tag = gtk_input_add_full(session->isdn_fd,
+					           GDK_INPUT_READ,
+					           gtk_handle_isdn_input,
+						   NULL,
+					           (gpointer) session,
+						   NULL);
 
   /* server functionality */
-  session->gdk_local_input_tag = gdk_input_add(session->local_sock,
-					       GDK_INPUT_READ,
-					       server_handle_local_input,
-					       (gpointer) session);
+  session->gtk_local_input_tag = gtk_input_add_full(session->local_sock,
+					            GDK_INPUT_READ,
+					            server_handle_local_input,
+						    NULL,
+					            (gpointer) session,
+						    NULL);
 }
 
 /*
@@ -511,10 +522,10 @@ void session_io_handlers_stop(session_t *session) {
   if (!(session->option_release_devices &&
 	(session->state == STATE_READY ||
 	 session->state == STATE_RINGING_QUIET))) {
-    gdk_input_remove(session->gdk_audio_input_tag);
+    gtk_input_remove(session->gtk_audio_input_tag);
   }
-  gdk_input_remove(session->gdk_isdn_input_tag);
-  gdk_input_remove(session->gdk_local_input_tag);
+  gtk_input_remove(session->gtk_isdn_input_tag);
+  gtk_input_remove(session->gtk_local_input_tag);
 }
 
 /*
@@ -699,6 +710,7 @@ void session_start_conversation(session_t *session) {
 				       session->record_checkbutton), FALSE);
 	}
 	free(digits);
+	cid_row_mark_record(session, session->cid_num - 1);
       } else {
 	fprintf(stderr, "Error generating audio filename.\n");
 	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(
@@ -719,27 +731,78 @@ void session_start_conversation(session_t *session) {
   }
 }
 
+/*
+ * to be called by a timeout and everyone who wants to stop a running effect
+ *   to go back to STATE_READY
+ */
+gint session_effect_stop_cb(gpointer data) {
+  session_t *session = (session_t *) data;
+
+  session_set_state(session, STATE_READY);
+  
+  return FALSE; /* don't call me regularly */
+}
+
 /*
  * To be called when we can write another block to sound device
  */
-void session_effect_callback(gpointer data, gint fd _U_,
-			     GdkInputCondition condition _U_) {
+void session_effect_callback(gpointer data,
+                             gint source _U_, GdkInputCondition condition _U_)
+{
   session_t *session = (session_t *) data;
-  int i;
+  int i = 0; /* position in raw byte output buffer */
+  int index_out = 0; /* logical position (sample number) in output buffer */
   unsigned char x; /* generated ulaw sample */
-  double seconds; /* position in seconds */
+  int just_read = 0;
+  int sndfile_buffer_index = 0;
+  double speed_factor = (double)session->effect_sndfile_buffer_frames /
+    (session->fragment_size_out / session->audio_sample_size_out);
+    /* factor from audio device to sndfile */
 
+  if (session->effect == EFFECT_SOUNDFILE) {
+    just_read = sf_readf_short(session->effect_sndfile,
+	session->effect_sndfile_buffer, session->effect_sndfile_buffer_frames);
+  }
+  
   /* fill buffer */
-  i = 0;
   while (i < session->fragment_size_out) {
-    seconds = (double)session->effect_pos / session->audio_speed_out;
-    x = fxgenerate(session, session->effect, 0, seconds);
+    if (session->effect == EFFECT_SOUNDFILE) {
+      double sum = 0.0;
+      double leftbound = speed_factor * index_out; /* sndfile bounds */
+      double rightbound = speed_factor * (index_out + 1);
+      
+      do {
+        int frame_sum = 0;
+	double weight = (rightbound < floor(leftbound + 1) ?
+	    rightbound : floor(leftbound + 1)) - leftbound;
+	  
+        sndfile_buffer_index = (int)leftbound;
+        if (sndfile_buffer_index < just_read) {
+          int t_count;
+          for (t_count=0; t_count < session->effect_sfinfo.channels; t_count++){
+	    frame_sum += (int) session->effect_sndfile_buffer[
+	      sndfile_buffer_index * session->effect_sfinfo.channels + t_count];
+          }
+          frame_sum /= session->effect_sfinfo.channels;
+        } else {
+	  frame_sum = 0;
+        }
+	sum += weight * frame_sum;
+	leftbound = floor(leftbound + 1);
+      } while (leftbound < rightbound);
+      sum /= speed_factor;
+      x = session->audio_LUT_generate[(int)sum / 256 + 128];
+    } else {
+      double seconds = (double)session->effect_pos / session->audio_speed_out;
+      x = fxgenerate(session, session->effect, 0, seconds);
+    }
     if (session->audio_sample_size_out == 1) {
       session->audio_outbuf[i++] = session->audio_LUT_in[(int)x];
     } else { /* audio_sample_size == 2 */
       session->audio_outbuf[i++] = session->audio_LUT_in[(int)x * 2];
       session->audio_outbuf[i++] = session->audio_LUT_in[(int)x * 2 + 1];
     }
+    index_out ++;
     session->effect_pos++;
   }
   
@@ -747,16 +810,50 @@ void session_effect_callback(gpointer data, gint fd _U_,
   write_buf(session->audio_fd_out,
 	    session->audio_outbuf,
 	    session->fragment_size_out);
+
+  if (session->effect == EFFECT_SOUNDFILE &&
+      just_read < session->effect_sndfile_buffer_frames &&
+      session->effect_filename)
+  {
+    gtk_timeout_add(
+      1000 * session->fragment_size_out / session->audio_sample_size_out *
+          (audio_get_write_fragments_total(session->audio_fd_out) -
+        audio_get_write_fragments_number(session->audio_fd_out)) /
+      session->audio_speed_out,
+      session_effect_stop_cb, session);
+    /* session->effect_filename provides a flag if we already want to stop */
+    free(session->effect_filename);
+    session->effect_filename = NULL;
+  }
 }
 
 /*
  * Start an effect (effect_t) playing on the sound device
+ *
+ * on kind == EFFECT_SOUNDFILE, session->effect_filename should be
+ * initialized so that session_effect_stop can free() it afterwards
  */
 void session_effect_start(session_t *session, enum effect_t kind) {
-  session->effect_tag = gdk_input_add(session->audio_fd_out,
-				      GDK_INPUT_WRITE,
-				      session_effect_callback,
-				      (gpointer) session);
+  if (kind == EFFECT_SOUNDFILE) {
+    session->effect_sfinfo.format = 0;
+    if (!(session->effect_sndfile =
+      sf_open(session->effect_filename, SFM_READ, &session->effect_sfinfo)))
+    {
+      fprintf(stderr, "Error on opening sound file.\n");
+    }
+    session->effect_sndfile_buffer_frames =
+      session->fragment_size_out / session->audio_sample_size_out *
+      session->effect_sfinfo.samplerate / session->audio_speed_out;
+    session->effect_sndfile_buffer = (short*) malloc (sizeof(short) *
+      session->effect_sndfile_buffer_frames * session->effect_sfinfo.channels);
+
+  }
+  session->effect_tag = gtk_input_add_full(session->audio_fd_out,
+				           GDK_INPUT_WRITE,
+				           session_effect_callback,
+					   NULL,
+				           (gpointer) session,
+					   NULL);
 
   session->effect = kind;
   session->effect_pos = 0;
@@ -767,7 +864,11 @@ void session_effect_start(session_t *session, enum effect_t kind) {
  */
 void session_effect_stop(session_t *session) {
   if (session->effect != EFFECT_NONE) { /* stop only if already playing */
-    gdk_input_remove(session->effect_tag);
+    if (session->effect == EFFECT_SOUNDFILE) {
+      sf_close(session->effect_sndfile);
+      free(session->effect_sndfile_buffer);
+    }
+    gtk_input_remove(session->effect_tag);
     session->effect = EFFECT_NONE;
   }
   session_io_handlers_stop(session);
@@ -840,12 +941,7 @@ int session_set_state(session_t *session, enum state_t state) {
     break;
   case STATE_RINGING:
     if (session->option_popup) {
-/* XXX: macros yet to define */
-#if GTK_MAJOR_VERSION >= 2
       gtk_window_present(GTK_WINDOW(session->main_window));
-#else
-      gdk_window_raise(session->main_window->window);
-#endif
     }
     if (session->effect == EFFECT_NONE)
       session_effect_start(session, EFFECT_RING);
@@ -853,11 +949,7 @@ int session_set_state(session_t *session, enum state_t state) {
     break;
   case STATE_RINGING_QUIET:
     if (session->option_popup) {
-#if GTK_MAJOR_VERSION >= 2
       gtk_window_present(GTK_WINDOW(session->main_window));
-#else
-      gdk_window_raise(session->main_window->window);
-#endif
     }
     if (debug) fprintf(stderr, "New state: STATE_RINGING_QUIET\n");
     break;
@@ -878,6 +970,11 @@ int session_set_state(session_t *session, enum state_t state) {
   case STATE_SERVICE:
     if (debug) fprintf(stderr, "New state: STATE_SERVICE\n");
     break;
+  case STATE_PLAYBACK:
+    if (session->effect == EFFECT_NONE)
+      session_effect_start(session, EFFECT_SOUNDFILE);
+    if (debug) fprintf(stderr, "New state: STATE_PLAYBACK\n");
+    break;
   default:
     fprintf(stderr, "Warning: session_set_state: Unhandled state.\n");
   }
@@ -917,13 +1014,41 @@ int session_set_state(session_t *session, enum state_t state) {
 }
 
 /*
- * callback for gdk on isdn input
+ * Callback: reinitialize isdn input watchdog
+ */
+static gboolean gtk_isdn_input_defer_timeout(session_t* session) {
+  session->gtk_isdn_input_tag = gtk_input_add_full(session->isdn_fd,
+					           GDK_INPUT_READ,
+					           gtk_handle_isdn_input,
+						   NULL,
+					           (gpointer) session,
+						   NULL);
+  return FALSE; /* don't call me regularly */
+}
+
+/* defer 300 milliseconds if appropriate */
+#define DEFER_INTERVAL 300
+
+/*
+ * put a graceful delay into GTK main loop to prevent permanent isdn input
+ * callback
+ */
+static void gtk_isdn_input_defer(session_t* session) {
+  gtk_input_remove(session->gtk_isdn_input_tag);
+  session->gtk_isdn_input_tag =
+    gtk_timeout_add(DEFER_INTERVAL,
+                    (GtkFunction) gtk_isdn_input_defer_timeout,
+		    session);
+}
+
+/*
+ * callback for gtk on isdn input
  *
  * input: data:      session (session_t *)
  *        fd:        file descriptor where we got the input from
  *        condition: will be GDK_INPUT_READ in this case
  */
-void gdk_handle_isdn_input(gpointer data, gint fd _U_,
+void gtk_handle_isdn_input(gpointer data, gint fd _U_,
 			   GdkInputCondition condition _U_) {
   session_t *session = (session_t *) data;
   char *temp;
@@ -1019,21 +1144,28 @@ void gdk_handle_isdn_input(gpointer data, gint fd _U_,
   case STATE_SERVICE:
     if (debug)
       fprintf(stderr, "Note: Got ISDN input in service mode.\n");
+    gtk_isdn_input_defer(session);
+    break;
+  case STATE_PLAYBACK:
+    if (debug)
+      fprintf(stderr, "Note: Got ISDN input in playback mode.\n");
+    gtk_isdn_input_defer(session);
     break;
   default:
     fprintf(stderr,
-	    "Warning: gdk_handle_isdn_input: Unknown session state.\n");
+	    "Warning: gtk_handle_isdn_input: Unknown session state.\n");
+    gtk_isdn_input_defer(session);
   }
 }
 
 /*
- * callback for gdk on audio isdn input
+ * callback for gtk on audio isdn input
  *
  * input: data:      session (session_t *)
  *        fd:        file descriptor where we got the input from
  *        condition: will be GDK_INPUT_READ in this case
  */
-void gdk_handle_audio_input(gpointer data, gint fd _U_,
+void gtk_handle_audio_input(gpointer data, gint fd _U_,
 			    GdkInputCondition condition _U_) {
   session_t *session = (session_t *) data;
 
@@ -1096,9 +1228,16 @@ void gdk_handle_audio_input(gpointer data, gint fd _U_,
     read(session->audio_fd_in,
 	 session->audio_inbuf, session->fragment_size_in);
     break;
+  case STATE_PLAYBACK:
+    if (debug > 1)
+      fprintf(stderr, "Warning: Got audio input in playback mode (ALSA?).\n");
+    /* flush audio input */
+    read(session->audio_fd_in,
+	 session->audio_inbuf, session->fragment_size_in);
+    break;
   default:
     fprintf(stderr,
-	    "Warning: gdk_handle_audio_input: Unknown session state.\n");
+	    "Warning: gtk_handle_audio_input: Unknown session state.\n");
     /* flush audio input */
     read(session->audio_fd_in,
 	 session->audio_inbuf, session->fragment_size_in);
@@ -1185,8 +1324,16 @@ void gtk_handle_pick_up_button(GtkWidget *widget _U_, gpointer data) {
     }
     break;
   case STATE_CONVERSATION: /* channel already working */
+    fprintf(stderr,
+	    "Non-sense warning: Pick up button pressed in conversation mode\n");
     break;
   case STATE_SERVICE:
+    fprintf(stderr,
+	    "Non-sense warning: Pick up button pressed in service mode\n");
+    break;
+  case STATE_PLAYBACK:
+    fprintf(stderr,
+	    "Non-sense warning: Pick up button pressed in playback mode\n");
     break;
   default:
     fprintf(stderr,
@@ -1228,6 +1375,11 @@ void gtk_handle_hang_up_button(GtkWidget *widget _U_, gpointer data) {
     cid_set_duration(session, NULL);
     break;
   case STATE_SERVICE:
+    fprintf(stderr,
+	    "Non-sense warning: Hang up button pressed in service mode\n");
+    break;
+  case STATE_PLAYBACK:
+    session_set_state(session, STATE_READY);
     break;
   default:
     fprintf(stderr,
